import logging
import os
import socket
import time

import colorlog

log_level = os.environ.get("PERRY_LOG_LEVEL", "INFO")
logger = logging.getLogger("perry")
logger.setLevel(getattr(logging, log_level))
logFormatter = colorlog.ColoredFormatter(
    fmt="%(log_color)s%(name)s :: %(levelname)-8s :: %(message)s"
)
handler = logging.StreamHandler()
handler.setFormatter(logFormatter)
logger.addHandler(handler)

def is_port_open(ip, port, timeout=2):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    result = sock.connect_ex((ip, port))
    return result == 0


def wait_until_port_is_open(ip, port, sleep_time=3, max_attempts=10):
    attempts = 0
    while not is_port_open(ip, port):
        attempts += 1
        if attempts >= max_attempts:
            raise RuntimeError(f"{ip}:{port} has not opened")
        time.sleep(sleep_time)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['perry_the_docker_agent']

package_data = \
{'': ['*'],
 'perry_the_docker_agent': ['sceptre/config/dev/*', 'sceptre/templates/*']}

install_requires = \
['PyYAML==5.4.1',
 'boto3==1.24.80',
 'click>=7.1.1,<7.2.0',
 'colorlog>=6.7.0,<7.0.0',
 'pathspec>=0.10.1,<0.11.0',
 'pydantic>=1.10.2,<2.0.0',
 'sceptre>=3.2.0,<4.0.0',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['perry = perry_the_docker_agent:main.app']}

setup_kwargs = {
    'name': 'perry-the-docker-agent',
    'version': '0.2.1',
    'description': 'Your cool remote docker agent in the cloud',
    'long_description': '## Perry The Docker Agent\n### Your cool remote docker agent in the cloud ⛄️\n\n[![PyPI version](https://badge.fury.io/py/perry-the-docker-agent.svg)](https://badge.fury.io/py/perry-the-docker-agent)\n![Python versions](https://img.shields.io/pypi/pyversions/perry-the-docker-agent.svg?style=flat-square&label=Python%20Versions)\n\nBased on [remote-docker-aws](https://github.com/lime-green/remote-docker-aws) 🙏🏾\n\nUse docker to develop services, but without the overhead of running docker on your machine! This is a development tool that you should use if your machine is low performance, or if you are running many docker services.\n\n### Why is this useful?\n\nFrees up your local machine for useful tasks\nsuch as running your code editor, browser, and email, leaving running Docker to a dedicated server instance.\nThe result is that your local machine functions faster, uses up less disk space, and consumes less power.\nMacOS users will also see noticeable speed improvements since Docker on Linux (which is\nwhat the remote hosts runs) is much more performant.\n\nThe downsides:\n- SSH tunnel communication is slower than local communication. However using an AWS region with low ping makes the latency unnoticeable. Find the region fastest for you using [this site](https://www.cloudping.info/)\n- Some more setup required to get everything configured properly and running (tunneling ports, syncing file changes)\n- Running the ec2 instance incurs an additional cost over running locally, although a t3.medium instance in Canada only costs just under 5 cents/hour\n\nHow it works: two processes are run, a sync and a tunnel process. \n- The sync process keeps local and remote files in sync so that the docker process run remotely can use docker volumes transparently\n- The tunnel process forwards ports needed so your local system can communicate with docker, plus additional ports as required, such as port 443 for browser communication\n\n## Setup\n1. First login to your AWS account and [create access keys to access AWS through the CLI](https://docs.aws.amazon.com/powershell/latest/userguide/pstools-appendix-sign-up.html)\n\n    You will need the following IAM policies:\n    - AmazonEC2FullAccess\n    - AWSCloudFormationFullAccess\n\n    And now in your terminal:\n\n    ```bash\n    # Replace dav with your name\n    # You will need to setup an AWS account if you don\'t have one\n    # and create access key credentials\n\n    aws configure --profile dav\n    export AWS_PROFILE=dav\n    ```\n\n1. Install pre-requisites\n\n   Have [Homebrew](https://brew.sh/) (Available on both macOS and Linux now!)\n\n   Have [pipx](https://github.com/pipxproject/pipx)\n\n    ```bash\n   pip install perry-the-docker-agent\n   # or.... if part of a project\n   poetry add perry-the-docker-agent\n\n   # Install unison sync utility\n   brew install unison\n\n   # Install file-watcher driver for unison\n   # On MacOS:\n   brew install autozimu/homebrew-formulas/unison-fsmonitor\n\n   # Or, on Linux since the above formula doesn\'t work:\n   brew install eugenmayer/dockersync/unox\n    ```\n\n1. Generate and upload a keypair to AWS\n\n    ```bash\n   perry create-key-pair\n    ```\n\n1. Create the ec2 instance\n\n    ```bash\n   perry create\n    ```\n\n## Daily Running\n\n1. Start the remote-docker ec2 instance\n    ```bash\n    perry start\n    ```\n   This will automatically switch the docker context for you. If you want to switch\n   back to the default agent run `docker context use default`\n\n1. In one terminal start the tunnel so that the ports you need to connect to are exposed\n    ```bash\n    perry tunnel\n    ```\n\n1. In another terminal sync file changes to the remote instance:\n    ```bash\n    perry sync\n    ```\n\n1. Develop and code! All services should be accessible and usable as usual (eg: `docker ps`, `docker-compose up`, etc.)\nas long as you are running `perry tunnel` and are forwarding the ports you need\n\n1. When you\'re done for the day don\'t forget to stop the instance to save money:\n    ```bash\n    perry stop\n    ```\n\n## Config File\nLooks for a config file at the path `./perry_config.yml` by default,\nwhich can be overriden by passing `--config-path`. \n\nAn example `perry_config.yml` file:\n```yaml\naws_region: af-south-1\ninstance_type: t3.large\nvolume_size: 100\nsync_dir: ./\nproject_id: "ranger"\nignore_dirs:\n  - .venv\n  - .git\n  - node_modules\n  - __pycache__\nremote_port_forwards:\n  local-webpack-app:\n    "8080": "8080"\nlocal_port_forwards:\n  user-api:\n    "2020": "2020"\n  blog-api:\n    "3030": "3030"\n```\n\n```bash\n Usage: perry [OPTIONS] COMMAND [ARGS]...                                                  \n                                                                                           \n╭─ Options ───────────────────────────────────────────────────────────────────────────────╮\n│ --config-path               TEXT  Path of the perry config                              │\n│                                   [default: ./perry_config.yml]                         │\n│ --install-completion              Install completion for the current shell.             │\n│ --show-completion                 Show completion for the current shell, to copy it or  │\n│                                   customize the installation.                           │\n│ --help                            Show this message and exit.                           │\n╰─────────────────────────────────────────────────────────────────────────────────────────╯\n╭─ Commands ──────────────────────────────────────────────────────────────────────────────╮\n│ create           Provision a new ec2 instance to use as the remote agent                │\n│ create-key-pair  Create and upload a new keypair to AWS for SSH access                  │\n│ delete           Delete the provisioned ec2 instance                                    │\n│ ssh              Connect to the remote agent via SSH                                    │\n│ start            Start the remote agent instance                                        │\n│ stop             Stop the remote agent instance                                         │\n│ sync             Sync the given directories with the remote instance                    │\n│ test             Test new config                                                        │\n│ tunnel           Create a SSH tunnel to the remote instance to connect with the docker  │\n│                  agent and containers                                                   │\n╰─────────────────────────────────────────────────────────────────────────────────────────╯\n```\n\nThe current configurable values can be seen in [config.py](perry_the_docker_agent/config.py)\n\n#### `aws_region` (takes precedence over `AWS_REGION` and `.aws/config`)\n- The region to create the instance in\n\n#### `instance_type`\n- Type of ec2 instance, defaults to: `t3.medium`\n\n#### `project_id`\n  - defaults to `None`\n  - Used to uniquely identify the instance, this is useful if multiple remote-docker agents\n  will be created in the same AWS account\n\n#### `key_path`\n  - defaults to: `~/.ssh/id_rsa_perry_{{project_id}}`\n\n#### `local_port_forwards`\n  - defaults to: `{}`\n  - Object containing label -> port mapping objects for opening the ports on the remote host.\n    A mapping of `"remote_port_forwards": {"my_app": {"80": "8080"}}` will open port 80 on your local machine\n    and point it to port 8080 of the remote-docker instance (which ostensibly a container is listening on).\n    The name doesn\'t do anything except help legibility.\n\n#### `remote_port_forwards`\n  - defaults to: `{}`\n  - Similar to `local_port_forwards` except will open the port on the remote instance.\n\n    This is useful to have frontend webpack apps accessible on the remote host\n\n#### `ignore_dirs`\n  - defaults to: `[]`\n  - list of directories to ignore\n\n#### `sync_dir`\n - directory to sync, will usually be the root fo the project\n\n#### `volume_size`\n - defaults to: `30` (GB)\n - Size of the ec2 volume.\n\n## Cost\nA t3.medium instance on ca-central-1 currently costs $0.046 /hour. [See current prices](https://aws.amazon.com/ec2/pricing/on-demand/)\n\nNothing else used should incur any cost with reasonable usage\n\n## Notes\n- See `perry --help` for more information on the commands available\n- The unison version running on the server and running locally have to\nmatch. If one of them updates to a newer version, you should update the other.\n\n## Perry improvements on [remote-docker-aws](https://github.com/lime-green/remote-docker-aws)\n- Perry is configured to the project by a local `perry_config.yml`\n- Perry uses [poetry](https://github.com/python-poetry/poetry), [pydantic](https://github.com/pydantic/pydantic) and [typer](https://github.com/tiangolo/typer)\n- Perry enables swap accounting on the remote ec2 instance which allows docker-compose resource limits\n- Perry uses a simpler method to sync and ignore directories.',
    'author': 'Daverin',
    'author_email': 'mail@daverin.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/daverin/perry-the-docker-agent/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import os
import shlex
import subprocess
from getpass import getuser
from typing import Dict, List

from .config import PerryConfig
from .constants import (
    INSTANCE_USERNAME
)
from .providers import AWSInstanceProvider, InstanceProvider
from .util import logger


class RemoteDockerClient:
    def __init__(
        self,
        instance: InstanceProvider,
        local_port_forwards: Dict[str, Dict[str, str]],
        remote_port_forwards: Dict[str, Dict[str, str]],
        ssh_key_path: str,
        sync_dir: str,
        ignore_dirs: str,
        project_code: str,
    ):
        self.instance = instance
        self.local_port_forwards = local_port_forwards
        self.remote_port_forwards = remote_port_forwards
        self.ssh_key_path = ssh_key_path
        self.sync_dir = sync_dir
        self.ignore_dirs = ignore_dirs
        self.project_code = project_code

    @classmethod
    def from_config(cls, config: PerryConfig):
        instance = AWSInstanceProvider(
            username=INSTANCE_USERNAME,
            project_code=config.project_code,
            aws_region=config.aws_region,
            instance_service_name=config.instance_service_name,
            instance_type=config.instance_type,
            instance_ami=config.instance_ami,
            ssh_key_pair_name=config.key_pair_name,
            volume_size=config.volume_size,
        )

        return cls(
            instance=instance,
            local_port_forwards=config.local_port_forwards,
            remote_port_forwards=config.remote_port_forwards,
            ssh_key_path=config.non_null_key_path,
            sync_dir=config.expanded_sync_dir,
            ignore_dirs=config.ignore_dirs,
            project_code=config.project_code
        )

    def get_ip(self) -> str:
        logger.debug("Retrieving IP address of instance")
        return self.instance.get_ip()

    def start_instance(self):
        logger.info("Starting instance")
        return self.instance.start_instance()

    def stop_instance(self):
        logger.info("Stopping instance")
        return self.instance.stop_instance()

    def enable_termination_protection(self):
        logger.info("Enabling Termination protection")
        return self.instance.enable_termination_protection()

    def disable_termination_protection(self):
        logger.info("Disabling Termination protection")
        return self.instance.disable_termination_protection()

    def is_termination_protection_enabled(self) -> bool:
        return self.instance.is_termination_protection_enabled()

    def start_tunnel(self):

        ip = self.instance.get_ip()
        cmd_s = (
            "sudo ssh -v -o ExitOnForwardFailure=yes -o StrictHostKeyChecking=no"
            " -o ServerAliveInterval=60 -N -T"
            f" -i {self.ssh_key_path} {self.instance.username}@{ip}"
        )

        target_sock = f"/var/run/{self.project_code}.sock"
        cmd_s += (
            f" -L {target_sock}:/var/run/docker.sock"
            " -o StreamLocalBindUnlink=yes"
            " -o PermitLocalCommand=yes"
            f" -o LocalCommand='sudo chown {getuser()} {target_sock}'"
        )

        for _name, port_mappings in self.local_port_forwards.items():
            for port_from, port_to in port_mappings.items():
                cmd_s += f" -L localhost:{port_from}:localhost:{port_to}"

        for _name, port_mappings in self.remote_port_forwards.items():
            for port_from, port_to in port_mappings.items():
                cmd_s += f" -R 0.0.0.0:{port_from}:localhost:{port_to}"

        logger.info("Starting tunnel")
        cmd = shlex.split(cmd_s)
        logger.debug("Running command: %s", cmd_s)

        logger.debug("Forwarding: ")
        logger.debug("Local: %s", self.local_port_forwards)
        logger.debug("Remote: %s", self.remote_port_forwards)
        subprocess.run(cmd, check=True)

    def create_instance(self):
        logger.info("Creating instance")
        return self.instance.create_instance(self.ssh_key_path)

    def delete_instance(self) -> Dict:
        logger.warning("Deleting instance")
        return self.instance.delete_instance()

    def ssh_connect(self, *, ssh_cmd: str = None, options: str = None):
        return self.instance.ssh_connect(
            ssh_key_path=self.ssh_key_path,
            ssh_cmd=ssh_cmd,
            options=options,
        )

    def ssh_run(self, *, ssh_cmd: str):
        return self.instance.ssh_run(
            ssh_key_path=self.ssh_key_path,
            ssh_cmd=ssh_cmd,
        )

    def create_keypair(self) -> Dict:
        return self.instance.create_keypair(self.ssh_key_path)

    def use_remote_context(self):
        logger.info(f"Switching docker context to {self.project_code}")

        subprocess.run(
            (
                f"docker context inspect {self.project_code} &>/dev/null || "
                "docker context create"
                f" --docker host=unix:///var/run/{self.project_code}.sock {self.project_code}"
            ),
            check=True,
            shell=True,
        )
        subprocess.run(
            f"docker context use {self.project_code} >/dev/null",
            check=True,
            shell=True,
        )

    def use_default_context(self):
        logger.info("Switching docker context to default")

        subprocess.run("docker context use default >/dev/null", check=True, shell=True)


    def _get_unison_cmd(
        self,
        *,
        ip: str,
        replica_path: str,
        ignore_dirs: List[str],
        force: bool = False,
        repeat_watch: bool = False,
    ) -> List[str]:
        cmd_s = (
            f"unison {replica_path}"
            f" 'ssh://{self.instance.username}@{ip}/{replica_path}'"
            f" -prefer {replica_path} -batch -sshargs '-i {self.ssh_key_path}'"
        )
        
        for ignore_dir in ignore_dirs:
            cmd_s += f" -ignore 'Name {{,.*,*,*/,.*/}}{ignore_dir}{{.*,*,*/,.*/}}'"

        if force:
            cmd_s += f" -force {replica_path}"
        if repeat_watch:
            cmd_s += " -repeat watch"

        return shlex.split(cmd_s.replace("\n", ""))

    def sync(self):
        ip = self.get_ip()

        logger.info("Ensuring remote directories exist")
        ssh_cmd_s = (
            f"sudo install -d -o {self.instance.username} -g {self.instance.username}"
        )
        ssh_cmd_s += f" -p {self.sync_dir}"
        
        self.ssh_run(ssh_cmd=ssh_cmd_s)

        logger.info("deleting any files owned by root (this messes with unison)")

        clean_cmd = (
            f"find {self.sync_dir} -user root | xargs sudo rm -rf  | echo 'no files to delete'"
        )
        self.ssh_run(ssh_cmd=clean_cmd)

        # First push the local replica's contents to remote
        logger.info("Pushing local files to remote server")

        subprocess.run(
            self._get_unison_cmd(
                ip=ip,
                replica_path=self.sync_dir,
                ignore_dirs=self.ignore_dirs,
                force=True,
            ),
            check=True,
        )

        # Then watch for update
        logger.info("Watching local and remote filesystems for changes")
        watch_cmd = self._get_unison_cmd(
            ip=ip,
            replica_path=self.sync_dir,
            ignore_dirs=self.ignore_dirs,
            repeat_watch=True,
        )

        logger.debug("Running command: %s", watch_cmd)
        os.execvp(watch_cmd[0], watch_cmd)


def create_remote_docker_client(
    config: PerryConfig,
) -> RemoteDockerClient:
    return RemoteDockerClient.from_config(config)
